from typing import Any, List

import torch.nn as nn
from sklearn.base import BaseEstimator

from .dummy import TorchDummyClassifier, TorchDummyRegressor
from .gradient_boosting import (
    TorchGradientBoostingClassifier,
    TorchGradientBoostingRegressor,
)
from .label_binarizer import TorchLabelBinarizer
from .linear_model import (
    TorchLinearClassifier,
    TorchLinearRegression,
    TorchLogisticRegression,
    TorchSGDClassifier,
)
from .nn import TorchMLPClassifier, TorchMLPRegressor
from .nystroem import TorchNystroem
from .pca import TorchPCA
from .pipeline import TorchPipeline
from .stacking import TorchStackingClassifier, TorchStackingRegressor
from .standard_scaler import TorchStandardScaler
from .svc import TorchSVC
from .svr import TorchSVR
from .tree import TorchDecisionTreeClassifier, TorchDecisionTreeRegressor
from .ttr import TorchTransformedTargetRegressor

# This list is intentionally kept alphabetical.
_REGISTRY = [
    TorchDecisionTreeClassifier,
    TorchDecisionTreeRegressor,
    TorchDummyClassifier,
    TorchDummyRegressor,
    TorchGradientBoostingClassifier,
    TorchGradientBoostingRegressor,
    TorchLabelBinarizer,
    TorchLinearClassifier,
    TorchLinearRegression,
    TorchLogisticRegression,
    TorchMLPClassifier,
    TorchMLPRegressor,
    TorchNystroem,
    TorchPCA,
    TorchPipeline,
    TorchSGDClassifier,
    TorchStackingClassifier,
    TorchStackingRegressor,
    TorchStandardScaler,
    TorchSVC,
    TorchSVR,
    TorchTransformedTargetRegressor,
]


def wrap(obj: BaseEstimator) -> nn.Module:
    for x in _REGISTRY:
        for cls in x.supported_classes():
            if isinstance(obj, cls):
                return x.wrap(obj)
    raise ValueError(f"unsupported sklearn estimator type: {type(obj)}")


def supported_classes() -> List[Any]:
    return [x for y in _REGISTRY for x in y.supported_classes()]



import numpy as np
import pytest
import torch
import torch.jit
from sklearn.svm import SVR

from .svr import TorchSVR


@pytest.mark.parametrize(
    ("kernel",),
    [
        ("linear",),
        ("rbf",),
    ],
)
def test_svr(kernel):
    x = np.random.random(size=(100, 2)) * 2 - 1
    y = np.sin(x[:, 0]) + np.cos(x[:, 1]) * 0.2

    sk_obj = SVR(kernel=kernel)
    sk_obj.fit(x, y)
    th_obj = torch.jit.script(TorchSVR.wrap(sk_obj))

    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_obj.predict(x)
        actual = th_obj.predict(x_th).numpy()
        assert (np.abs(expected - actual) < 1e-5).all()



import numpy as np
import pytest
import torch
import torch.jit
from sklearn.ensemble import GradientBoostingClassifier, GradientBoostingRegressor

from .gradient_boosting import (
    TorchGradientBoostingClassifier,
    TorchGradientBoostingRegressor,
)


@pytest.mark.parametrize(
    ("target_type", "loss", "init_zero"),
    [
        ("binary", "exponential", False),
        ("binary", "exponential", True),
        ("binary", "deviance", False),
        ("binary", "deviance", True),
        ("multi", "deviance", False),
        ("multi", "deviance", True),
    ],
)
def test_gradient_boosting_classifier(target_type, loss, init_zero):
    xs = np.random.normal(size=(1000, 3)) * 2
    noised = xs + np.random.normal(size=xs.shape) * 0.1
    if target_type == "binary":
        ys = np.mean(noised, axis=1) < 0
    else:
        # Non-uniformly spaced or centered class labels.
        ys = np.round(np.clip(np.mean(noised, axis=1), -3, 3)).astype(np.int32) * 2

    sk_obj = GradientBoostingClassifier(
        loss=loss, init="zero" if init_zero else None, n_estimators=5
    )
    sk_obj.fit(xs, ys)
    th_obj = torch.jit.script(TorchGradientBoostingClassifier.wrap(sk_obj))
    xs_th = torch.from_numpy(xs)

    with torch.no_grad():
        expected = sk_obj.decision_function(xs)
        actual = th_obj.decision_function(xs_th).numpy()
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)

        expected = sk_obj.predict_proba(xs)
        actual = th_obj.predict_proba(xs_th).numpy()
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)

        expected = sk_obj.predict(xs)
        actual = th_obj.predict(xs_th).numpy()
        assert actual.shape == expected.shape
        assert (actual == expected).all()


@pytest.mark.parametrize(("init_zero",), [(False,), (True,)])
def test_gradient_boosting_regressor(init_zero):
    xs = np.random.normal(size=(1000, 3)) * 2
    ys = np.mean(xs, axis=-1)
    sk_obj = GradientBoostingRegressor(
        init="zero" if init_zero else None, n_estimators=5
    )
    sk_obj.fit(xs, ys)
    th_obj = torch.jit.script(TorchGradientBoostingRegressor.wrap(sk_obj))
    xs_th = torch.from_numpy(xs)

    with torch.no_grad():
        expected = sk_obj.predict(xs)
        actual = th_obj.predict(xs_th).numpy()
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)



from typing import List, Type, Union

import torch
import torch.jit
import torch.nn as nn
import torch.nn.functional as F
from sklearn.neural_network import MLPClassifier, MLPRegressor

from .label_binarizer import TorchLabelBinarizer


class _WrappedMLP(nn.Module):
    def __init__(self, layers: nn.Sequential, out_act: str):
        super().__init__()
        assert out_act in ["identity", "softmax", "logistic"]
        self.layers = layers
        self.out_act = out_act

    @classmethod
    def wrap(cls, obj: Union[MLPClassifier, MLPRegressor]) -> "_WrappedMLP":
        acts = {
            "identity": nn.Identity,
            "logistic": nn.Sigmoid,
            "tanh": nn.Tanh,
            "relu": nn.ReLU,
        }
        if obj.activation not in acts:
            raise ValueError(f"unsupported activation: {obj.activation}")
        act = acts[obj.activation]
        modules = []
        for weights, biases in zip(obj.coefs_, obj.intercepts_):
            w = torch.from_numpy(weights)
            b = torch.from_numpy(biases)
            module = nn.Linear(weights.shape[0], weights.shape[1], dtype=w.dtype)
            with torch.no_grad():
                module.weight.copy_(w.t())
                module.bias.copy_(b)
            modules.append(module)
            modules.append(act())
        del modules[-1]  # output activation handled separately.
        return cls(layers=nn.Sequential(*modules), out_act=obj.out_activation_)

    def forward(self, x: torch.Tensor, include_negative: bool = False) -> torch.Tensor:
        x = self.layers(x)
        if x.shape[1] == 1 and self.out_act == "logistic" and include_negative:
            x = torch.cat([F.logsigmoid(-x), F.logsigmoid(x)], dim=-1)
        elif self.out_act == "logistic":
            x = F.logsigmoid(x)
        elif self.out_act == "softmax":
            x = F.log_softmax(x, dim=-1)
        return x


class TorchMLPClassifier(nn.Module):
    def __init__(
        self,
        module: _WrappedMLP,
        label_binarizer: TorchLabelBinarizer,
    ):
        super().__init__()
        self.module = module
        self.label_binarizer = label_binarizer

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [MLPClassifier]

    @classmethod
    def wrap(cls, obj: MLPClassifier) -> "TorchMLPClassifier":
        return cls(
            module=_WrappedMLP.wrap(obj),
            label_binarizer=TorchLabelBinarizer.wrap(obj._label_binarizer),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        probs = self.module(x).exp()
        return self.label_binarizer.inverse_transform(probs)

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict_log_proba(x).exp()

    @torch.jit.export
    def predict_log_proba(self, x: torch.Tensor) -> torch.Tensor:
        return self.module(x, include_negative=True)


class TorchMLPRegressor(nn.Module):
    def __init__(
        self,
        module: _WrappedMLP,
    ):
        super().__init__()
        self.module = module

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [MLPRegressor]

    @classmethod
    def wrap(cls, obj: MLPRegressor) -> "TorchMLPRegressor":
        return cls(module=_WrappedMLP.wrap(obj))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        out = self.module(x)
        if out.shape[1] == 1:
            return out.view(-1)
        return out



from typing import List, Optional, Type

import torch
import torch.jit
import torch.nn as nn
from sklearn.decomposition import PCA


class TorchPCA(nn.Module):
    def __init__(
        self,
        mean: torch.Tensor,
        components: torch.Tensor,
        scale: Optional[torch.Tensor],
    ):
        super().__init__()
        self.mean = nn.Parameter(mean)
        self.components = nn.Parameter(components)
        if scale is not None:
            self.scale = nn.Parameter(scale)
        else:
            self.scale = 1.0

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [PCA]

    @classmethod
    def wrap(cls, obj: PCA) -> "TorchPCA":
        components = torch.from_numpy(obj.components_)
        mean = obj.mean_
        if mean is None:
            mean = torch.zeros(
                components.shape[1], device=components.device, dtype=components.dtype
            )
        else:
            mean = torch.from_numpy(mean)
        explained_variance = torch.from_numpy(obj.explained_variance_)
        return cls(
            mean=mean,
            components=components,
            scale=explained_variance.rsqrt() if obj.whiten else None,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.transform(x)

    @torch.jit.export
    def transform(self, x: torch.Tensor) -> torch.Tensor:
        return ((x - self.mean) @ self.components.t()) * self.scale

    @torch.jit.export
    def inverse_transform(self, x: torch.Tensor) -> torch.Tensor:
        return ((x / self.scale) @ self.components) + self.mean



import numpy as np
import torch
import torch.jit
from sklearn.linear_model import SGDClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from .linear_model_test import xor_and_dataset
from .pipeline import TorchPipeline


def test_pipeline_classifier():
    x, y = xor_and_dataset()
    sk_obj = Pipeline(
        [("scale", StandardScaler()), ("classify", SGDClassifier(loss="log"))]
    )
    sk_obj.fit(x, y)
    th_obj = torch.jit.script(TorchPipeline.wrap(sk_obj))

    x_th = torch.from_numpy(x).clone()
    with torch.no_grad():
        expected = sk_obj.predict(x)
        actual = th_obj.predict(x_th).numpy()
        assert (expected == actual).all()

        # forward() should be equivalent
        actual = th_obj(x_th).numpy()
        assert (expected == actual).all()

        expected = sk_obj.decision_function(x)
        actual = th_obj.decision_function(x_th).numpy()
        assert np.allclose(expected, actual)

        expected = sk_obj.predict_proba(x)
        actual = th_obj.predict_proba(x_th).numpy()
        assert np.allclose(expected, actual)

        expected = sk_obj.predict_log_proba(x)
        actual = th_obj.predict_log_proba(x_th).numpy()
        assert np.allclose(expected, actual)


def test_pipeline_transform():
    x, y = xor_and_dataset()
    sk_obj = Pipeline(
        [
            ("center", StandardScaler(with_std=False)),
            ("scale", StandardScaler()),
        ]
    )
    sk_obj.fit(x, y)
    th_obj = torch.jit.script(TorchPipeline.wrap(sk_obj))

    x_th = torch.from_numpy(x)
    with torch.no_grad():
        expected = sk_obj.transform(x)
        actual = th_obj.transform(x_th).numpy()
        assert (expected == actual).all()

        # forward() should be equivalent
        actual = th_obj(x_th).numpy()
        assert (expected == actual).all()

        expected = sk_obj.inverse_transform(x)
        actual = th_obj.inverse_transform(x_th).numpy()
        assert np.allclose(expected, actual)



from typing import List, Type

import torch
import torch.jit
import torch.nn as nn
from sklearn.kernel_approximation import Nystroem

from .kernel import Kernel


class TorchNystroem(nn.Module):
    def __init__(
        self, kernel: Kernel, components: torch.Tensor, normalization: torch.Tensor
    ):
        super().__init__()
        self.kernel = kernel
        self.components = nn.Parameter(components)
        self.normalization = nn.Parameter(normalization)

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [Nystroem]

    @classmethod
    def wrap(cls, obj: Nystroem) -> "TorchNystroem":
        kernel = Kernel.wrap(obj)
        return cls(
            kernel=kernel,
            components=torch.from_numpy(obj.components_),
            normalization=torch.from_numpy(obj.normalization_),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.transform(x)

    @torch.jit.export
    def transform(self, x: torch.Tensor) -> torch.Tensor:
        return self.kernel(x, self.components) @ self.normalization.t()



from abc import abstractmethod
from copy import deepcopy
from typing import List, Type, Union

import numpy as np
import torch
import torch.jit
import torch.nn as nn
import torch.nn.functional as F
from sklearn.linear_model import (
    LinearRegression,
    LogisticRegression,
    LogisticRegressionCV,
    Ridge,
    RidgeClassifier,
    RidgeClassifierCV,
    RidgeCV,
    SGDClassifier,
    SGDRegressor,
)
from sklearn.linear_model._base import LinearModel
from sklearn.svm import LinearSVC, LinearSVR


class TorchLinearModel(nn.Module):
    def __init__(
        self,
        model: LinearModel,
    ):
        super().__init__()
        if hasattr(model, "densify"):
            model = deepcopy(model)
            model.densify()
        else:
            assert isinstance(
                model.coef_, np.ndarray
            ), "sparse linear model is not supported"
        weights = torch.from_numpy(model.coef_)
        biases = torch.from_numpy(np.array(model.intercept_))
        if len(weights.shape) == 1:
            weights = weights[None]
        self.weights = nn.Parameter(weights)
        self.biases = nn.Parameter(biases)

    def _decision_function(self, x: torch.Tensor) -> torch.Tensor:
        outputs = (x @ self.weights.t()) + self.biases
        if outputs.shape[1] == 1:
            return outputs.view(-1)
        return outputs


class TorchLinearRegression(TorchLinearModel):
    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [LinearRegression, Ridge, RidgeCV, SGDRegressor, LinearSVR]

    @classmethod
    def wrap(
        cls, obj: Union[LinearRegression, Ridge, RidgeCV, SGDRegressor, LinearSVR]
    ) -> "TorchLinearRegression":
        return cls(obj)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Predict class labels for the given feature vectors.

        An alias for self.predict().
        """
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Predict class labels for the given feature vectors.
        """
        return self._decision_function(x)

    def _decision_function(self, x: torch.Tensor) -> torch.Tensor:
        outputs = (x @ self.weights.t()) + self.biases
        if outputs.shape[1] == 1:
            return outputs.view(-1)
        return outputs


class TorchLinearClassifier(TorchLinearModel):
    def __init__(self, model: LinearModel):
        super().__init__(model=model)
        self.register_buffer("classes", torch.from_numpy(model.classes_))

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [RidgeClassifier, RidgeClassifierCV, LinearSVC]

    @classmethod
    def wrap(
        cls, obj: Union[RidgeClassifier, RidgeClassifierCV, LinearSVC]
    ) -> "TorchLinearClassifier":
        return cls(model=obj)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Predict class labels for the given feature vectors.

        An alias for self.predict().
        """
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Predict class labels for the given feature vectors.
        """
        scores = self.decision_function(x)
        if len(scores.shape) == 1:
            indices = (scores > 0).long()
        else:
            indices = scores.argmax(-1)
        return self.classes[indices]

    @torch.jit.export
    def decision_function(self, x: torch.Tensor) -> torch.Tensor:
        return self._decision_function(x)


class TorchSGDClassifier(TorchLinearClassifier):
    def __init__(
        self,
        loss: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.loss = loss

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [SGDClassifier]

    @classmethod
    def wrap(cls, obj: SGDClassifier) -> "TorchSGDClassifier":
        return cls(
            loss=obj.loss,
            model=obj,
        )

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict_log_proba(x).exp()

    @torch.jit.export
    def predict_log_proba(self, x: torch.Tensor) -> torch.Tensor:
        if self.loss == "log":
            logits = self.decision_function(x)
            if len(logits.shape) == 1:
                return torch.stack(
                    [F.logsigmoid(-logits), F.logsigmoid(logits)], dim=-1
                )
            # This is a one-versus-rest classifier.
            return F.log_softmax(F.logsigmoid(logits), dim=-1)
        else:
            raise RuntimeError(
                "probability prediction not supported for loss: " + self.loss
            )


class TorchLogisticRegression(TorchLinearClassifier):
    def __init__(
        self,
        multi_class: str,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.multi_class = multi_class

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [LogisticRegression, LogisticRegressionCV]

    @classmethod
    def wrap(
        cls, obj: Union[LogisticRegression, LogisticRegressionCV]
    ) -> "TorchLogisticRegression":
        multi_class = obj.multi_class
        if multi_class == "auto" and (
            len(obj.classes_) == 2 or obj.solver == "liblinear"
        ):
            multi_class = "ovr"
        return cls(multi_class=multi_class, model=obj)

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict_log_proba(x).exp()

    @torch.jit.export
    def predict_log_proba(self, x: torch.Tensor) -> torch.Tensor:
        logits = self.decision_function(x)
        if self.multi_class == "ovr":
            if len(logits.shape) == 1:
                return torch.stack(
                    [F.logsigmoid(-logits), F.logsigmoid(logits)], dim=-1
                )
            return F.log_softmax(F.logsigmoid(logits), dim=-1)
        else:
            if len(logits.shape) == 1:
                logits = torch.stack([-logits, logits], dim=-1)
            return F.log_softmax(logits, dim=-1)



import torch
import torch.nn as nn


def fill_unsupported(module: nn.Module, *names: str):
    """
    Fill unsupported method names on the module with a function that raises an
    exception when it is called. This can be used to appease TorchScript.
    """
    for name in names:
        if not hasattr(module, name):

            def unsupported_fn(
                _: torch.Tensor, unsup_method_name: str = name
            ) -> torch.Tensor:
                raise RuntimeError(
                    f"method {unsup_method_name} is not supported on this object"
                )

            setattr(module, name, unsupported_fn)



from typing import Any, List, Optional, Type, Union

import numpy as np
import torch
import torch.jit
import torch.nn as nn
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor


class BaseTree(nn.Module):
    def __init__(
        self,
        tree: Any,
    ):
        super().__init__()

        # Track which nodes are branches/leaves. This will be used to index
        # matrices to avoid storing unused values.
        is_branch = np.zeros(len(tree.value), dtype=bool)

        # Rows are leaves and columns are branches in the next two arrays.
        # The conditions bitmap indicates, for a given leaf, what branches
        # must be true.
        # The mask bitmap indicates, for a given leaf, what branches must be
        # checked (i.e. ignore branches outside of a leaf's path).
        node_conditions = np.zeros([len(tree.value)] * 2, dtype=bool)
        node_conditions_mask = node_conditions.copy()

        # Cache of the used nodes for each branch.
        # This is just node_conditions_mask with the diagonal set to True.
        decision_paths = node_conditions.copy()

        def enumerate_tree(node_id: int, parent_id: Optional[int] = None):
            if parent_id is not None:
                node_conditions_mask[node_id] = node_conditions_mask[parent_id]
                node_conditions_mask[node_id][parent_id] = True
                decision_paths[node_id] = decision_paths[parent_id]
            decision_paths[node_id, node_id] = True
            left_id, right_id = (
                tree.children_left[node_id],
                tree.children_right[node_id],
            )
            if left_id != right_id:
                is_branch[node_id] = True
                node_conditions[left_id] = node_conditions[node_id]
                node_conditions[right_id] = node_conditions[node_id]
                node_conditions[right_id][node_id] = True
                enumerate_tree(left_id, node_id)
                enumerate_tree(right_id, node_id)

        enumerate_tree(0)

        # We will perform accumulations (sometimes via matmul) along each row
        # of these matrices, so we must be able to store the largest sum.
        max_depth = np.max(np.sum(node_conditions_mask.astype(np.int64), axis=-1))
        if max_depth < 2 ** 7:
            mat_dtype = torch.int8
        elif max_depth < 2 ** 15:
            mat_dtype = torch.int16
        else:
            mat_dtype = torch.int32

        is_leaf = np.logical_not(is_branch)
        self.register_buffer("feature", torch.from_numpy(tree.feature[is_branch]))
        self.value = nn.Parameter(torch.from_numpy(tree.value[is_leaf]))
        self.threshold = nn.Parameter(torch.from_numpy(tree.threshold[is_branch]))
        self.register_buffer(
            "cond",
            torch.from_numpy(node_conditions[np.ix_(is_leaf, is_branch)]).to(mat_dtype),
        )
        self.register_buffer(
            "cond_mask",
            torch.from_numpy(node_conditions_mask[np.ix_(is_leaf, is_branch)]).to(
                mat_dtype
            ),
        )
        self.register_buffer(
            "decision_paths", torch.from_numpy(decision_paths[is_leaf])
        )

    def _leaf_indices(self, x: torch.Tensor) -> torch.Tensor:
        comparisons = (x[:, self.feature] > self.threshold).to(self.cond)
        cond_counts = comparisons @ self.cond.t()
        # Two conditions:
        # 1. no_false_neg: every condition that needs to be true is true.
        # 2. no_false_pos: no other conditions along the decision path are true.
        no_false_neg = cond_counts == self.cond.sum(1)
        no_false_pos = cond_counts == (comparisons @ self.cond_mask.t())
        return (no_false_neg & no_false_pos).int().argmax(-1)

    @torch.jit.export
    def raw_values(self, x: torch.Tensor) -> torch.Tensor:
        return self.value[self._leaf_indices(x)]

    @torch.jit.export
    def decision_path(self, x: torch.Tensor):
        return self.decision_paths[self._leaf_indices(x)]


class TorchDecisionTreeRegressor(BaseTree):
    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [DecisionTreeRegressor]

    @classmethod
    def wrap(cls, obj: DecisionTreeRegressor) -> "TorchDecisionTreeRegressor":
        return TorchDecisionTreeRegressor(obj.tree_)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        y = self.raw_values(x).view(len(x), -1)
        if y.shape[1] == 1:
            y = y.view(-1)
        return y


class TorchDecisionTreeClassifier(BaseTree):
    def __init__(
        self,
        n_outputs: int,
        n_classes: Union[int, List[int]],
        classes: Union[List[torch.Tensor], torch.Tensor],
        tree: Any,
    ):
        super().__init__(tree)
        self.n_outputs = int(n_outputs)
        if n_outputs == 1:
            self.outputs = nn.ModuleList([_SingleClassOutput(n_classes, classes)])
        else:
            self.outputs = nn.ModuleList(
                [_SingleClassOutput(x, y) for x, y in zip(n_classes, classes)]
            )

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [DecisionTreeClassifier]

    @classmethod
    def wrap(cls, obj: DecisionTreeClassifier) -> "TorchDecisionTreeClassifier":
        if obj.n_outputs_ == 1:
            classes = torch.from_numpy(obj.classes_)
        else:
            classes = [torch.from_numpy(x) for x in obj.classes_]
        return TorchDecisionTreeClassifier(
            obj.n_outputs_, obj.n_classes_, classes, obj.tree_
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        y = self.raw_values(x)
        outputs = []
        for i, x in enumerate(self.outputs):
            outputs.append(x(y[:, i]))
        if self.n_outputs == 1:
            return outputs[0]
        else:
            return torch.stack(outputs, dim=-1)

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> Union[torch.Tensor, List[torch.Tensor]]:
        y = self.raw_values(x)
        outputs = []
        for i, x in enumerate(self.outputs):
            outputs.append(x.predict_proba(y[:, i]))
        return self._collapse(outputs)

    @torch.jit.export
    def predict_log_proba(
        self, x: torch.Tensor
    ) -> Union[torch.Tensor, List[torch.Tensor]]:
        y = self.raw_values(x)
        outputs = []
        for i, x in enumerate(self.outputs):
            outputs.append(x.predict_log_proba(y[:, i]))
        return self._collapse(outputs)

    def _collapse(
        self, x: List[torch.Tensor]
    ) -> Union[torch.Tensor, List[torch.Tensor]]:
        if self.n_outputs == 1:
            return x[0]
        else:
            return x


class _SingleClassOutput(nn.Module):
    def __init__(self, n_classes: int, classes: torch.Tensor):
        super().__init__()
        self.n_classes = int(n_classes)
        self.register_buffer("classes", classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.classes[x[:, : self.n_classes].argmax(1)]

    def predict_proba(self, x: torch.Tensor) -> torch.Tensor:
        proba = x[:, : self.n_classes]
        normalizer = proba.sum(1, keepdim=True)
        proba = proba / torch.where(normalizer == 0.0, 1.0, normalizer)
        return proba

    def predict_log_proba(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict_proba(x).log()



from typing import List, Type

import torch
import torch.jit
import torch.nn as nn
from sklearn.ensemble import StackingClassifier, StackingRegressor

from .util import fill_unsupported


class TorchStackingClassifier(nn.Module):
    def __init__(
        self,
        passthrough: bool,
        estimators: List[nn.Module],
        stack_methods: List[str],
        final_estimator: nn.Module,
        classes: torch.Tensor,
    ):
        super().__init__()
        self.passthrough = passthrough
        self.estimators = nn.ModuleList(estimators)
        for model in self.estimators:
            fill_unsupported(model, "predict_proba", "decision_function", "predict")
        self.stack_methods = stack_methods
        self.final_estimator = final_estimator
        fill_unsupported(
            self.final_estimator, "predict_proba", "decision_function", "predict"
        )
        self.register_buffer("classes", classes)

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [StackingClassifier]

    @classmethod
    def wrap(cls, obj: StackingClassifier) -> "TorchStackingClassifier":
        from .wrap import wrap

        return cls(
            passthrough=obj.passthrough,
            estimators=[wrap(x) for x in obj.estimators_],
            stack_methods=obj.stack_method_,
            final_estimator=wrap(obj.final_estimator_),
            classes=torch.from_numpy(obj.classes_),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        return self.classes[self.final_estimator.predict(self.transform(x))]

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> torch.Tensor:
        return self.final_estimator.predict_proba(self.transform(x))

    @torch.jit.export
    def decision_function(self, x: torch.Tensor) -> torch.Tensor:
        return self.final_estimator.decision_function(self.transform(x))

    @torch.jit.export
    def transform(self, x: torch.Tensor) -> torch.Tensor:
        outputs = []
        for i, estimator in enumerate(self.estimators):
            method = self.stack_methods[i]
            if method == "predict":
                out = estimator.predict(x)
            elif method == "predict_proba":
                out = estimator.predict_proba(x)
                if out.shape[1] == 2:
                    out = out[:, 1:]
            else:
                assert method == "decision_function"
                out = estimator.decision_function(x)
            if len(out.shape) == 1:
                out = out[:, None]
            outputs.append(out)
        if self.passthrough:
            outputs.append(x.view(len(x), -1))
        return torch.cat(outputs, dim=-1).to(x)


class TorchStackingRegressor(nn.Module):
    def __init__(
        self,
        passthrough: bool,
        estimators: List[nn.Module],
        final_estimator: nn.Module,
    ):
        super().__init__()
        self.passthrough = passthrough
        self.estimators = nn.ModuleList(estimators)
        self.final_estimator = final_estimator

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [StackingRegressor]

    @classmethod
    def wrap(cls, obj: StackingRegressor) -> "TorchStackingRegressor":
        from .wrap import wrap

        return cls(
            passthrough=obj.passthrough,
            estimators=[wrap(x) for x in obj.estimators_],
            final_estimator=wrap(obj.final_estimator_),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        return self.final_estimator.predict(self.transform(x))

    @torch.jit.export
    def transform(self, x: torch.Tensor) -> torch.Tensor:
        outputs = []
        for i, estimator in enumerate(self.estimators):
            outputs.append(estimator.predict(x).view(-1, 1))
        if self.passthrough:
            outputs.append(x.view(len(x), -1))
        return torch.cat(outputs, dim=-1).to(x)



# type: ignore

import numpy as np
import pytest
import torch
import torch.jit
from sklearn.decomposition import PCA

from .pca import TorchPCA


@pytest.mark.parametrize(("whiten",), [(False,), (True,)])
def test_pca(whiten: bool):
    rng = np.random.RandomState(1337)
    xs = (rng.normal(size=(1000, 4)) @ rng.normal(size=(4, 10))) + rng.normal(
        size=(10,)
    )
    sk_obj = PCA(n_components=3, whiten=whiten)
    sk_obj.fit(xs)
    th_obj = torch.jit.script(TorchPCA.wrap(sk_obj))

    xs_th = torch.from_numpy(xs)

    with torch.no_grad():
        actual = th_obj.transform(xs_th).numpy()
        expected = sk_obj.transform(xs)
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)

        ys = expected
        actual = th_obj.inverse_transform(torch.from_numpy(ys)).numpy()
        expected = sk_obj.inverse_transform(ys)
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)



from typing import List, Tuple, Type

import torch
import torch.jit
import torch.nn as nn
from sklearn.pipeline import Pipeline

from .util import fill_unsupported


class TorchPipeline(nn.Module):
    def __init__(self, stages: List[Tuple[str, nn.Module]]):
        super().__init__()
        self.transforms = nn.ModuleDict({k: v for k, v in stages[:-1]})
        for transform in self.transforms.values():
            fill_unsupported(transform, "inverse_transform")

        self.final_name, self.final = stages[-1]
        fill_unsupported(
            self.final,
            "decision_function",
            "predict",
            "predict_proba",
            "predict_log_proba",
            "transform",
            "inverse_transform",
        )

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [Pipeline]

    @classmethod
    def wrap(cls, obj: Pipeline) -> "TorchPipeline":
        from .wrap import wrap

        mapping = []
        for k, v in obj.steps:
            mapping.append((k, wrap(v)))
        return cls(mapping)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Run the pipeline and call forward() on the final model.
        """
        return self.final(self._run_transforms(x))

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Run the pipeline and call predict() on the final model.
        """
        return self.final.predict(self._run_transforms(x))

    @torch.jit.export
    def decision_function(self, x: torch.Tensor) -> torch.Tensor:
        """
        Run the pipeline and call decision_function() on the final model.
        """
        return self.final.decision_function(self._run_transforms(x))

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> torch.Tensor:
        """
        Run the pipeline and call predict_proba() on the final model.
        """
        return self.final.predict_proba(self._run_transforms(x))

    @torch.jit.export
    def predict_log_proba(self, x: torch.Tensor) -> torch.Tensor:
        """
        Run the pipeline and call predict_log_proba() on the final model.
        """
        return self.final.predict_log_proba(self._run_transforms(x))

    @torch.jit.export
    def transform(self, x: torch.Tensor) -> torch.Tensor:
        """
        Run the pipeline and call transform() on the final model.
        """
        return self.final.transform(self._run_transforms(x))

    def _run_transforms(self, x: torch.Tensor) -> torch.Tensor:
        for transform in self.transforms.values():
            x = transform(x)
        return x

    @torch.jit.export
    def inverse_transform(self, x: torch.Tensor) -> torch.Tensor:
        """
        Run the pipeline and call inverse_transform() on the final model.
        """
        x = self.final.inverse_transform(x)
        for transform in self.transforms.values()[::-1]:
            x = transform.inverse_transform(x)
        return x



import warnings
from typing import Tuple

import numpy as np
import pytest
import torch
import torch.jit
from sklearn.exceptions import ConvergenceWarning
from sklearn.neural_network import MLPClassifier, MLPRegressor

from .nn import TorchMLPClassifier, TorchMLPRegressor


@pytest.mark.parametrize(
    ("activation", "class_type", "hidden_sizes"),
    [
        ("relu", "binary", (15, 20)),
        ("relu", "multiclass", (15, 20)),
        ("relu", "multilabel-indicator", (15, 20)),
        ("tanh", "multiclass", (15, 20)),
        ("logistic", "multiclass", (15, 20)),
        ("identity", "multiclass", (15, 20)),
        ("tanh", "multiclass", (15)),
        ("tanh", "multiclass", (30, 30, 30)),
    ],
)
def test_mlp_classifier(
    activation: str, class_type: str, hidden_sizes: Tuple[int, ...]
):
    rng = np.random.RandomState(1337)
    x = rng.normal(size=(1000, 10))
    perturbed = x + rng.normal(size=x.shape)
    if class_type == "binary":
        y = perturbed.sum(1) > 0
    elif class_type == "multiclass":
        y = perturbed.argmax(1)
    elif class_type == "multilabel-indicator":
        y = perturbed > 0

    sk_model = MLPClassifier(
        hidden_layer_sizes=hidden_sizes,
        activation=activation,
    )
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ConvergenceWarning)
        sk_model.fit(x[:-100], y[:-100])
    th_model = torch.jit.script(TorchMLPClassifier.wrap(sk_model))

    x = x[-100:]
    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_model.predict_proba(x)
        actual = th_model.predict_proba(x_th).numpy()
        assert expected.shape == actual.shape
        assert (np.abs(actual - expected) < 1e-3).all()

        expected = sk_model.predict(x)
        actual = th_model.predict(x_th).numpy()
        assert expected.shape == actual.shape
        assert (actual == expected).all()


@pytest.mark.parametrize(
    ("activation", "output_2d", "hidden_sizes"),
    [
        ("relu", False, (15, 20)),
        ("tanh", False, (15, 20)),
        ("logistic", False, (15, 20)),
        ("identity", False, (15, 20)),
        ("relu", True, (15, 20)),
    ],
)
def test_mlp_regressor(activation: str, output_2d: bool, hidden_sizes: Tuple[int, ...]):
    rng = np.random.RandomState(1337)
    x = rng.normal(size=(1000, 10))
    perturbed = x + rng.normal(size=x.shape)
    if not output_2d:
        y = perturbed.sum(1)
    else:
        y = np.stack([perturbed.sum(1), np.sqrt(np.abs(perturbed.sum(1)))], axis=-1)

    sk_model = MLPRegressor(
        hidden_layer_sizes=hidden_sizes,
        activation=activation,
    )
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ConvergenceWarning)
        sk_model.fit(x[:-100], y[:-100])
    th_model = torch.jit.script(TorchMLPRegressor.wrap(sk_model))

    x = x[-100:]
    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_model.predict(x)
        actual = th_model.predict(x_th).numpy()
        assert expected.shape == actual.shape
        assert np.allclose(actual, expected)



import warnings
from typing import Callable, Tuple, Union

import numpy as np
import pytest
import torch
import torch.jit
from sklearn.datasets import load_breast_cancer, load_digits
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import (
    LinearRegression,
    LogisticRegression,
    LogisticRegressionCV,
    Ridge,
    RidgeClassifier,
    RidgeClassifierCV,
    RidgeCV,
    SGDClassifier,
    SGDRegressor,
)
from sklearn.svm import LinearSVC, LinearSVR

from .linear_model import (
    TorchLinearClassifier,
    TorchLinearRegression,
    TorchLogisticRegression,
    TorchSGDClassifier,
)
from .svc_test import quadrant_dataset


def noisy_binary_mean(**_):
    """A dataset with a simple linear decision function but noisy labels"""
    xs = np.random.normal(size=(1000, 10))
    ys = np.mean(xs + 0.1 + np.random.normal(size=xs.shape) * 0.1, axis=-1) < 0
    return xs, ys.astype(np.int32)


def xor_dataset(**_) -> Tuple[np.ndarray, np.ndarray]:
    """
    A binary classification problem that is not linearly separable.
    """
    x = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0]])
    y = np.array([0.0, 1.0, 1.0, 0.0])
    return np.tile(x, [10, 1]), np.tile(y, [10])


def xor_and_dataset(**_) -> Tuple[np.ndarray, np.ndarray]:
    """
    A three-class problem that is not linearly separable.
    """
    x = np.array([[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0]])
    y = np.array([0, 1, 1, 2])
    return np.tile(x, [10, 1]), np.tile(y, [10])


@pytest.mark.parametrize(
    (
        "sk_obj",
        "num_targets",
    ),
    [
        (cls(fit_intercept=fit_intercept), num_targets)
        for cls in [LinearRegression, Ridge, RidgeCV]
        for fit_intercept in [False, True]
        for num_targets in [1, 2]
    ]
    + [(SGDRegressor(fit_intercept=False), 1), (SGDRegressor(fit_intercept=True), 1)],
)
def test_linear_regression(
    sk_obj: Union[LinearRegression, Ridge, RidgeCV], num_targets: int
):
    x = np.random.random(size=(100, 2)) * 2 - 1
    if num_targets == 1:
        y = np.sin(x[:, 0]) + np.cos(x[:, 1]) * 0.2
    else:
        assert num_targets == 2
        y = np.stack(
            [
                np.sin(x[:, 0]) + np.cos(x[:, 1]) * 0.2,
                np.cos(x[:, 0]) + np.sin(x[:, 1]) * 0.2,
            ],
            axis=1,
        )

    sk_obj.fit(x, y)
    th_obj = torch.jit.script(TorchLinearRegression.wrap(sk_obj))

    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_obj.predict(x)
        actual = th_obj.predict(x_th).numpy()
        assert actual.shape == expected.shape
        assert (np.abs(expected - actual) < 1e-5).all()


@pytest.mark.parametrize(
    ("fit_intercept",),
    [
        (False,),
        (True,),
    ],
)
def test_linear_svr(fit_intercept: bool):
    x = np.random.random(size=(100, 2)) * 2 - 1
    y = np.sin(x[:, 0]) + np.cos(x[:, 1]) * 0.2

    sk_obj = LinearSVR(
        fit_intercept=fit_intercept, intercept_scaling=2.314, max_iter=10000
    )
    sk_obj.fit(x, y)
    assert LinearSVR in TorchLinearRegression.supported_classes()
    th_obj = torch.jit.script(TorchLinearRegression.wrap(sk_obj))

    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_obj.predict(x)
        actual = th_obj.predict(x_th).numpy()
        assert (np.abs(expected - actual) < 1e-5).all()


@pytest.mark.parametrize(
    ("dataset", "loss", "check_probs", "fit_intercept", "space_classes"),
    [
        (load_breast_cancer, "log", False, True, False),
        (load_breast_cancer, "log", False, True, True),
        (load_digits, "log", False, True, False),
        (load_digits, "log", False, True, True),
        (xor_dataset, "log", True, True, False),
        (xor_and_dataset, "log", True, True, False),
        (xor_and_dataset, "log", True, False, False),
        (xor_and_dataset, "hinge", False, True, False),
    ],
)
def test_sgd_classifier(
    dataset: Callable[..., Tuple[np.ndarray, np.ndarray]],
    loss: str,
    check_probs: bool,
    fit_intercept: bool,
    space_classes: bool,
):
    x, y = dataset(return_X_y=True)
    if space_classes:
        n_classes = np.max(y)
        y = np.where(y > n_classes // 2, 1 + (n_classes - y) + n_classes // 2, y)

    sk_obj = SGDClassifier(loss=loss, fit_intercept=fit_intercept)
    sk_obj.fit(x, y)
    th_obj = torch.jit.script(TorchSGDClassifier.wrap(sk_obj))

    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_obj.decision_function(x)
        actual = th_obj.decision_function(x_th).numpy()
        assert np.allclose(expected, actual)

        expected = sk_obj.predict(x)
        actual = th_obj(x_th).numpy()
        assert (expected == actual).all()

        if check_probs:
            expected = sk_obj.predict_log_proba(x)
            actual = th_obj.predict_log_proba(x_th).numpy()
            assert not np.isnan(expected).any()
            assert np.allclose(np.exp(expected), np.exp(actual))

            expected = sk_obj.predict_proba(x)
            actual = th_obj.predict_proba(x_th).numpy()
            assert np.isfinite(expected).any()
            assert np.isfinite(actual).any()
            assert np.allclose(expected, actual)


@pytest.mark.parametrize(
    ("dataset", "fit_intercept", "space_classes", "multi_class", "cv"),
    [
        (load_breast_cancer, True, False, "auto", False),
        (load_breast_cancer, True, True, "auto", False),
        (load_digits, True, False, "auto", False),
        (load_digits, True, True, "auto", False),
        (xor_dataset, True, False, "auto", False),
        (xor_dataset, True, False, "multinomial", False),
        (xor_dataset, True, False, "ovr", False),
        (xor_and_dataset, True, False, "auto", False),
        (xor_and_dataset, True, False, "ovr", False),
        (xor_and_dataset, True, False, "multinomial", False),
        (xor_and_dataset, False, False, "auto", False),
        (noisy_binary_mean, True, True, "ovr", False),
        (noisy_binary_mean, True, True, "multinomial", False),
        (noisy_binary_mean, True, True, "auto", True),
    ],
)
def test_logistic_regression(
    dataset: Callable[..., Tuple[np.ndarray, np.ndarray]],
    fit_intercept: bool,
    space_classes: bool,
    multi_class: str,
    cv: bool,
):
    x, y = dataset(return_X_y=True)
    if space_classes:
        n_classes = np.max(y)
        y = np.where(y > n_classes // 2, 1 + (n_classes - y) + n_classes // 2, y)

    sk_obj = (LogisticRegression if not cv else LogisticRegressionCV)(
        fit_intercept=fit_intercept, multi_class=multi_class
    )
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ConvergenceWarning)
        sk_obj.fit(x, y)
    th_obj = torch.jit.script(TorchLogisticRegression.wrap(sk_obj))

    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_obj.decision_function(x)
        actual = th_obj.decision_function(x_th).numpy()
        assert actual.shape == expected.shape
        assert np.allclose(expected, actual)

        expected = sk_obj.predict(x)
        actual = th_obj(x_th).numpy()
        assert actual.shape == expected.shape
        assert (expected == actual).all()

        expected = sk_obj.predict_proba(x)
        actual = th_obj.predict_proba(x_th).numpy()
        assert actual.shape == expected.shape
        assert np.isfinite(expected).any()
        assert np.isfinite(actual).any()
        assert np.allclose(expected, actual)

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning)
            expected = sk_obj.predict_log_proba(x)
        actual = th_obj.predict_log_proba(x_th).numpy()
        assert actual.shape == expected.shape
        assert not np.isnan(expected).any()
        assert np.allclose(np.exp(expected), np.exp(actual))


@pytest.mark.parametrize(
    ("dataset", "fit_intercept", "space_classes", "cv"),
    [
        (load_breast_cancer, True, False, False),
        (load_breast_cancer, True, True, False),
        (load_breast_cancer, True, True, True),
        (load_breast_cancer, False, True, True),
        (load_digits, True, False, False),
        (load_digits, True, True, False),
        (load_digits, True, True, True),
        (load_digits, False, True, True),
    ],
)
def test_ridge_classifier(
    dataset: Callable[..., Tuple[np.ndarray, np.ndarray]],
    fit_intercept: bool,
    space_classes: bool,
    cv: bool,
):
    x, y = dataset(return_X_y=True)
    if space_classes:
        n_classes = np.max(y)
        y = np.where(y > n_classes // 2, 1 + (n_classes - y) + n_classes // 2, y)

    sk_obj = (RidgeClassifier if not cv else RidgeClassifierCV)(
        fit_intercept=fit_intercept
    )
    sk_obj.fit(x, y)
    th_obj = torch.jit.script(TorchLinearClassifier.wrap(sk_obj))

    x_th = torch.from_numpy(x)

    with torch.no_grad():
        expected = sk_obj.decision_function(x)
        actual = th_obj.decision_function(x_th).numpy()
        assert actual.shape == expected.shape
        assert np.allclose(expected, actual)

        expected = sk_obj.predict(x)
        actual = th_obj(x_th).numpy()
        assert actual.shape == expected.shape
        assert (expected == actual).all()


@pytest.mark.parametrize(
    ("fit_intercept", "n_classes", "space_classes"),
    [
        (False, 2, False),
        (True, 2, False),
        (False, 3, False),
        (True, 3, False),
        (True, 4, True),
    ],
)
def test_linear_svc(fit_intercept: bool, n_classes: int, space_classes: int):
    xs, ys = quadrant_dataset(n_classes, space_classes)

    test_xs = np.random.random(size=(128, 2)) * 2 - 1
    test_xs_th = torch.from_numpy(test_xs)

    model = LinearSVC(fit_intercept=fit_intercept, intercept_scaling=2.314)
    model.fit(xs, ys)
    model_th = torch.jit.script(TorchLinearClassifier.wrap(model))

    with torch.no_grad():
        expected = model.decision_function(test_xs)
        actual = model_th.decision_function(test_xs_th).numpy()
        assert (np.abs(actual - expected) < 1e-8).all()

        expected = model.predict(test_xs)
        actual = model_th.predict(test_xs_th).numpy()
        assert (actual == expected).all()



from typing import List, Type, Union

import numpy as np
import torch
import torch.jit
import torch.nn as nn
from sklearn.dummy import DummyClassifier, DummyRegressor


class TorchDummyClassifier(nn.Module):
    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [DummyClassifier]

    @classmethod
    def wrap(
        cls, obj: DummyClassifier
    ) -> Union["TorchDummyClassifierSingle", "TorchDummyClassifierMulti"]:
        assert not obj.sparse_output_, "sparse classifiers are not supported"
        if isinstance(obj.n_classes_, list):
            return TorchDummyClassifierMulti(
                singles=[
                    TorchDummyClassifierSingle(
                        classes=torch.from_numpy(obj.classes_[i]),
                        class_prior=torch.from_numpy(obj.class_prior_[i]),
                        strategy=obj.strategy,
                        constant=(
                            torch.from_numpy(np.array(obj.constant[i]))
                            if obj.constant is not None
                            else torch.from_numpy(obj.classes_[i])[0]
                        ),
                    )
                    for i in range(len(obj.n_classes_))
                ],
            )
        else:
            return TorchDummyClassifierSingle(
                classes=torch.from_numpy(obj.classes_),
                class_prior=torch.from_numpy(obj.class_prior_),
                strategy=obj.strategy,
                constant=(
                    torch.from_numpy(np.array(obj.constant))
                    if obj.constant is not None
                    else torch.from_numpy(obj.classes_)[0]
                ),
            )


class TorchDummyClassifierSingle(TorchDummyClassifier):
    def __init__(
        self,
        classes: torch.Tensor,
        class_prior: torch.Tensor,
        strategy: str,
        constant: torch.Tensor,
    ):
        super().__init__()
        assert strategy in [
            "most_frequent",
            "prior",
            "stratified",
            "uniform",
            "constant",
        ]
        self.register_buffer("classes", classes)
        self.class_prior = nn.Parameter(class_prior)
        self.strategy = strategy
        self.register_buffer("constant", constant)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        if self.strategy == "most_frequent" or self.strategy == "prior":
            return self.classes[self.class_prior.argmax()].repeat(len(x))
        elif self.strategy == "stratified":
            return self.classes[
                torch.multinomial(
                    self.class_prior, num_samples=len(x), replacement=True
                )
            ]
        elif self.strategy == "uniform":
            return self.classes[
                torch.randint(low=0, high=len(self.classes), size=(len(x),))
            ]
        return self.constant.view(1).repeat(len(x))

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> torch.Tensor:
        if self.strategy == "most_frequent":
            out = self.classes == self.classes[self.class_prior.argmax()]
            return out[None].repeat(len(x), 1).to(self.class_prior)
        elif self.strategy == "prior":
            return self.class_prior[None].repeat(len(x), 1)
        elif self.strategy == "stratified":
            samples = torch.multinomial(
                self.class_prior, num_samples=len(x), replacement=True
            )
            return (self.classes == self.classes[samples[:, None]]).to(self.class_prior)
        elif self.strategy == "uniform":
            out = torch.ones_like(self.class_prior) / len(self.class_prior)
            return out[None].repeat(len(x), 1)
        out = self.classes == self.constant
        return out[None].repeat(len(x), 1).to(self.class_prior)

    @torch.jit.export
    def predict_log_proba(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict_proba(x).log()


class TorchDummyClassifierMulti(TorchDummyClassifier):
    def __init__(self, singles: List[TorchDummyClassifierSingle]):
        super().__init__()
        self.singles = nn.ModuleList(singles)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        return torch.stack([single.predict(x) for single in self.singles], dim=1)

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> List[torch.Tensor]:
        return [single.predict_proba(x) for single in self.singles]

    @torch.jit.export
    def predict_log_proba(self, x: torch.Tensor) -> torch.Tensor:
        return [single.predict_log_proba(x) for single in self.singles]


class TorchDummyRegressor(nn.Module):
    def __init__(self, strategy: str, constant: torch.Tensor):
        super().__init__()
        if strategy == "constant":
            self.register_buffer("constant", constant)
        else:
            self.constant = nn.Parameter(constant)

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [DummyRegressor]

    @classmethod
    def wrap(cls, obj: DummyRegressor) -> "TorchDummyRegressor":
        return cls(
            strategy=obj.strategy, constant=torch.from_numpy(obj.constant_).view(-1)
        )

    def forward(self, x: torch.Tensor):
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        res = self.constant[None].repeat(len(x), 1)
        if res.shape[1] == 1:
            res = res.view(-1)
        return res



from typing import List, Optional, Tuple, Type, Union

import torch
import torch.jit
import torch.nn as nn
import torch.nn.functional as F
from sklearn.svm import SVC, NuSVC

from .kernel import Kernel


class TorchSVC(nn.Module):
    def __init__(
        self,
        kernel: Kernel,
        ovr: bool,
        break_ties: bool,
        n_support: List[int],
        support_vectors: torch.Tensor,
        intercept: torch.Tensor,
        dual_coef: torch.Tensor,
        classes: torch.Tensor,
        prob_a: Optional[torch.Tensor],
        prob_b: Optional[torch.Tensor],
    ):
        super().__init__()
        self.kernel = kernel
        self.ovr = ovr
        self.break_ties = break_ties
        self.n_support = n_support
        self.support_vectors = nn.Parameter(support_vectors)
        self.intercept = nn.Parameter(intercept)
        self.dual_coef = nn.Parameter(dual_coef)
        self.register_buffer("classes", classes)
        self.supports_prob = prob_a is not None and prob_b is not None
        self.prob_a = nn.Parameter(
            torch.ones_like(intercept) if prob_a is None else prob_a
        )
        self.prob_b = nn.Parameter(
            torch.zeros_like(intercept) if prob_b is None else prob_b
        )
        self.n_classes = len(n_support)

        self.sv_offsets = []
        offset = 0
        for count in n_support:
            self.sv_offsets.append(offset)
            offset += count

        self._ovo_index_map = [-1 for _ in range(self.n_classes ** 2)]
        k = 0
        for i in range(self.n_classes - 1):
            for j in range(i + 1, self.n_classes):
                self._ovo_index_map[i * self.n_classes + j] = k
                self._ovo_index_map[j * self.n_classes + i] = k
                k += 1

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [SVC, NuSVC]

    @classmethod
    def wrap(cls, obj: Union[SVC, NuSVC]) -> "TorchSVC":
        assert not obj._sparse, "sparse SVC not supported"
        assert obj.decision_function_shape in ["ovo", "ovr"]
        return cls(
            kernel=Kernel.wrap(obj),
            ovr=obj.decision_function_shape == "ovr",
            break_ties=obj.break_ties,
            n_support=obj.n_support_.tolist(),
            support_vectors=torch.from_numpy(obj.support_vectors_),
            intercept=torch.from_numpy(obj.intercept_),
            dual_coef=torch.from_numpy(obj.dual_coef_),
            classes=torch.from_numpy(obj.classes_),
            prob_a=torch.from_numpy(obj.probA_) if obj.probability else None,
            prob_b=torch.from_numpy(obj.probB_) if obj.probability else None,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.predict(x)

    @torch.jit.export
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        ovo, ovr = self.decision_function_ovo_ovr(x)
        if self.n_classes == 2:
            indices = (ovo.view(-1) > 0).long()
        elif self.ovr and self.break_ties:
            indices = ovr.argmax(dim=-1)
        else:
            indices = ovr.round().argmax(dim=-1)
        return self.classes[indices]

    @torch.jit.export
    def decision_function(self, x: torch.Tensor) -> torch.Tensor:
        ovo, ovr = self.decision_function_ovo_ovr(x)
        if len(self.n_support) == 2:
            return ovo.view(-1)
        elif self.ovr:
            return ovr
        else:
            return ovo

    @torch.jit.export
    def predict_log_proba(self, x: torch.Tensor) -> torch.Tensor:
        assert self.supports_prob, "model must be trained with probability=True"
        if self.n_classes == 2:
            ovo, _ = self.decision_function_ovo_ovr(x)
            logit = ovo * self.prob_a - self.prob_b
            return torch.cat([F.logsigmoid(logit), F.logsigmoid(-logit)], dim=-1)
        return self.predict_proba(x).log()

    @torch.jit.export
    def predict_proba(self, x: torch.Tensor) -> torch.Tensor:
        assert self.supports_prob, "model must be trained with probability=True"
        ovo, _ = self.decision_function_ovo_ovr(x)
        if self.n_classes == 2:
            # This shortcut optimization is present in the latest LibSVM
            # but not in the scikit-learn fork. As a result, the scikit
            # version is slightly less accurate in the binary case.
            # https://github.com/scikit-learn/scikit-learn/blob/82df48934eba1df9a1ed3be98aaace8eada59e6e/sklearn/svm/src/libsvm/svm.cpp#L2925

            # For some reason, prob_a has the opposite sign for this case.
            logit = ovo * self.prob_a - self.prob_b
            return torch.cat([logit.sigmoid(), (-logit).sigmoid()], dim=-1)

        probs = (-ovo * self.prob_a - self.prob_b).sigmoid()
        min_prob = 1e-7
        probs = probs.clamp(min_prob, 1 - min_prob)

        matrix = self._prob_matrix(probs)
        inv_diag = 1 / (torch.diagonal(matrix, dim1=1, dim2=2) + 1e-12)
        guess = torch.ones(len(x), self.n_classes).to(matrix) / self.n_classes
        masks = torch.eye(self.n_classes).to(matrix)

        # We must use variable name `i` and pre-define `delta`
        # to appease the TorchScript compiler. For the loop, the
        # variable name "_" does not work.
        delta = torch.zeros_like(guess)
        for i in range(max(100, self.n_classes)):
            # Coordinate descent is performed for each dimension between
            # checks of the stopping criterion:
            # https://github.com/cjlin1/libsvm/blob/eedbb147ea79af44f2ecdca1064f2c6a8fe6462d/svm.cpp#L1873-L1883
            for coord in range(self.n_classes):
                mask = masks[coord]
                # Eq 47 from https://www.csie.ntu.edu.tw/~cjlin/papers/libsvm.pdf
                mg = (matrix @ guess[:, :, None]).view(guess.shape)
                outer = torch.einsum("ij,ij->i", guess, mg)[:, None]
                delta = outer - mg

                guess = guess + inv_diag * delta * mask
                guess = guess / guess.sum(dim=-1, keepdim=True)

            # Stopping criterion.
            if delta.abs().max().item() < 0.005 / self.n_classes:
                break
        return guess

    def _prob_matrix(self, probs: torch.Tensor) -> torch.Tensor:
        """
        Compute the pairwise probability matrix Q from page 31 of
        https://www.csie.ntu.edu.tw/~cjlin/papers/libsvm.pdf.
        """
        matrix_elems = []
        for i in range(self.n_classes):
            for j in range(self.n_classes):
                if i == j:
                    prob_sum = torch.zeros_like(probs[:, 0])
                    for k in range(self.n_classes):
                        if k != i:
                            prob_sum = prob_sum + self._pairwise_prob(probs, k, i) ** 2
                    matrix_elems.append(prob_sum)
                else:
                    matrix_elems.append(
                        -self._pairwise_prob(probs, i, j)
                        * self._pairwise_prob(probs, j, i)
                    )
        return torch.stack(matrix_elems, dim=-1).view(
            len(probs), self.n_classes, self.n_classes
        )

    def _pairwise_prob(self, probs: torch.Tensor, i: int, j: int) -> torch.Tensor:
        p = probs[:, self._ovo_index_map[i * self.n_classes + j]]
        if i > j:
            return 1 - p
        return p

    def decision_function_ovo_ovr(
        self, x: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Compute the one-versus-one and one-versus-rest decision functions."""
        kernel_out = self.kernel(x, self.support_vectors)

        votes = torch.zeros((len(x), self.n_classes)).to(kernel_out)
        confidence_sum = [
            torch.zeros(len(x)).to(kernel_out) for i in range(self.n_classes)
        ]
        ovos = []
        k = 0
        for i in range(self.n_classes - 1):
            for j in range(i + 1, self.n_classes):
                neg_confidence = (
                    self._dual_sum(kernel_out, i, j)
                    + self._dual_sum(kernel_out, j, i)
                    + self.intercept[k]
                )
                k += 1
                ovos.append(neg_confidence)
                confidence_sum[i] = confidence_sum[i] + neg_confidence
                confidence_sum[j] = confidence_sum[j] - neg_confidence
                pred = neg_confidence < 0
                votes[:, i] += torch.logical_not(pred)
                votes[:, j] += pred
        ovo = torch.stack(ovos, dim=-1)
        confidences = torch.stack(confidence_sum, dim=-1)

        # Don't overrule votes while allowing confidences to break ties.
        # https://github.com/scikit-learn/scikit-learn/blob/7e1e6d09bcc2eaeba98f7e737aac2ac782f0e5f1/sklearn/utils/multiclass.py#L475
        confidences = confidences / (3 * (confidences.abs() + 1))

        return ovo, votes + confidences

    def _dual_sum(self, kernel_out: torch.Tensor, i: int, j: int) -> torch.Tensor:
        assert j != i
        if j > i:
            j -= 1
        offset, count = self.sv_offsets[i], self.n_support[i]
        coeffs = self.dual_coef[j, offset : offset + count]
        kernel_row = kernel_out[:, offset : offset + count]
        return kernel_row @ coeffs



from typing import List, Optional, Type

import torch
import torch.jit
import torch.nn as nn
from sklearn.preprocessing import LabelBinarizer


class TorchLabelBinarizer(nn.Module):
    def __init__(
        self,
        classes: torch.Tensor,
        neg_label: int,
        pos_label: int,
        y_type: str,
    ):
        super().__init__()
        self.register_buffer("classes", classes)
        self.neg_label = neg_label
        self.pos_label = pos_label
        self.y_type = y_type

    @classmethod
    def supported_classes(cls) -> List[Type]:
        return [LabelBinarizer]

    @classmethod
    def wrap(cls, obj: LabelBinarizer) -> "TorchLabelBinarizer":
        assert obj.y_type_ in ["multiclass", "multilabel-indicator", "binary"]
        return cls(
            classes=torch.from_numpy(obj.classes_),
            pos_label=int(obj.pos_label),
            neg_label=int(obj.neg_label),
            y_type=obj.y_type_,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.transform(x)

    @torch.jit.export
    def transform(self, x: torch.Tensor) -> torch.Tensor:
        if self.y_type == "multilabel-indicator":
            # The logic in the sklearn code is pretty inconsistent and broken
            # when using arbitrary pos_label/neg_label, but we replicate its
            # behavior for this case as closely as possible.
            # https://github.com/scikit-learn/scikit-learn/blob/7e1e6d09bcc2eaeba98f7e737aac2ac782f0e5f1/sklearn/preprocessing/_label.py#L549-L566
            pos_label = self.pos_label
            pos_switch = pos_label == 0
            if pos_switch:
                pos_label = -self.neg_label
            if pos_label != 1:
                # We preserve x's dtype here, which means pos_label will be
                # cast to it. For example, if x is a bool, and pos_label > 1,
                # then pos_label will be forced to 1.
                # This is a bug in the scikit-learn implementation, so we
                # immitate it.
                x = torch.where(x != 0, torch.tensor(pos_label).to(x), x)
            x = x.long()
            if self.neg_label != 0:
                x = torch.where(x == 0, self.neg_label, x)
            if pos_switch:
                x = torch.where(x == pos_label, 0, x)
            return x
        elif self.y_type == "multiclass":
            return torch.where(
                x[..., None] == self.classes, self.pos_label, self.neg_label
            ).long()
        else:
            assert self.y_type == "binary"
            return torch.where(
                x[..., None] == self.classes[1:], self.pos_label, self.neg_label
            ).long()

    @torch.jit.export
    def inverse_transform(
        self, x: torch.Tensor, threshold: Optional[float] = None
    ) -> torch.Tensor:
        if self.y_type == "multiclass":
            return self.classes[x.argmax(-1)]
        if threshold is None:
            threshold = (self.pos_label + self.neg_label) / 2
        outputs = self.classes[(x > threshold).long()]
        if self.y_type == "binary":
            outputs = outputs.view(-1)
        return outputs



import warnings

import numpy as np
import pytest
import torch
import torch.jit
from sklearn.dummy import DummyClassifier, DummyRegressor
from sklearn.ensemble import StackingClassifier, StackingRegressor
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import SGDClassifier
from sklearn.svm import LinearSVC, LinearSVR

from .stacking import TorchStackingClassifier, TorchStackingRegressor


@pytest.mark.parametrize(
    ("drop_classifier", "binary", "passthrough", "method"),
    [
        (drop_classifier, binary, passthrough, method)
        for drop_classifier in [False, True]
        for binary in [False, True]
        for passthrough in [False, True]
        for method in ["auto", "predict_proba", "decision_function", "predict"]
        if not passthrough or (method == "auto" and binary == False)
    ],
)
def test_stacking_classifier(
    drop_classifier: bool, binary: bool, passthrough: bool, method: str
):
    rs = np.random.RandomState(1338)
    xs = rs.normal(size=(1000, 10))
    noised_xs = xs + rs.normal(size=xs.shape) * 0.1
    if binary:
        ys = (np.mean(noised_xs, axis=-1) > 0).astype(np.int32) * 2 + 1
    else:
        float_val = np.sum(noised_xs, axis=-1) / np.sqrt(xs.shape[1])
        ys = np.round(np.clip(float_val, -3, 3)).astype(np.int32) * 2

    classifiers = [
        ("dummy", DummyClassifier(strategy="prior")),  # no decision_function
        ("sgd_classifier", SGDClassifier(loss="log")),  # all methods supported
        ("linear_svc", LinearSVC()),  # no predict_proba
    ]
    if method == "decision_function":
        del classifiers[0]
    elif method == "predict_proba":
        del classifiers[2]

    sk_obj = StackingClassifier(
        estimators=classifiers,
        stack_method=method,
        passthrough=passthrough,
    )
    if drop_classifier:
        sk_obj.set_params(sgd_classifier="drop")
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ConvergenceWarning)
        warnings.filterwarnings("ignore", category=UserWarning)
        sk_obj.fit(xs, ys)

    th_obj = torch.jit.script(TorchStackingClassifier.wrap(sk_obj))

    xs_th = torch.from_numpy(xs)
    with torch.no_grad():
        actual = th_obj.transform(xs_th).numpy()
        expected = sk_obj.transform(xs)
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)

        actual = th_obj.predict(xs_th).numpy()
        expected = sk_obj.predict(xs)
        assert actual.shape == expected.shape
        assert (actual == expected).all()

        actual = th_obj.decision_function(xs_th).numpy()
        expected = sk_obj.decision_function(xs)
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)

        actual = th_obj.predict_proba(xs_th).numpy()
        expected = sk_obj.predict_proba(xs)
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)


@pytest.mark.parametrize(
    ("drop_model", "passthrough"),
    [
        (drop_model, passthrough)
        for drop_model in [False, True]
        for passthrough in [False, True]
    ],
)
def test_stacking_regressor(drop_model: bool, passthrough: bool):
    rs = np.random.RandomState(1338)
    xs = rs.normal(size=(1000, 10))
    ys = np.mean((xs + rs.normal(size=xs.shape) * 0.1 + 0.1) ** 2, axis=-1)

    classifiers = [
        ("dummy", DummyRegressor()),
        ("linear_svr", LinearSVR()),
    ]

    sk_obj = StackingRegressor(
        estimators=classifiers,
        passthrough=passthrough,
    )
    if drop_model:
        sk_obj.set_params(linear_svr="drop")
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=ConvergenceWarning)
        sk_obj.fit(xs, ys)

    th_obj = torch.jit.script(TorchStackingRegressor.wrap(sk_obj))

    xs_th = torch.from_numpy(xs)
    with torch.no_grad():
        actual = th_obj.transform(xs_th).numpy()
        expected = sk_obj.transform(xs)
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)

        actual = th_obj.predict(xs_th).numpy()
        expected = sk_obj.predict(xs)
        assert actual.shape == expected.shape
        assert np.allclose(actual, expected)
